# -*- coding: utf-8 -*-
import json
import os.path
from typing import Any, Dict

from logzero import logger
import requests

from chaosiq.config import get_token
from chaosiq.exceptions import ConfigurationTokenMissing

__all__ = ["call_chaosiq_api"]

CHAOSIQ_API_URL = "https://api.chaosiq.io"


def call_chaosiq_api(path: str, payload: Dict[str, Any],
                     config: Dict[str, Any],
                     expected_status: int=200) -> requests.Response:
    """
    Perform a call against chaosiq.

    Probably do not bet on this function from outside this package, it will
    likely evolve.
    """
    try:
        token = get_token(config)
    except ConfigurationTokenMissing as err:
        logger.warn(str(err))
        return

    headers = {
        "Accept": "application/json",
        "Authorization": "Bearer {t}".format(t=token),
        "Connection": "close"
    }

    url = "{u}{p}".format(u=CHAOSIQ_API_URL, p=path)
    try:
        r = requests.post(
            url, json=payload, headers=headers, timeout=(2, 30))
    except requests.ConnectTimeout as c:
        logger.warn(
            "Failed to contact the ChaosIQ API service. Try again soon!")
    except requests.ConnectionError as e:
        logger.warn(
            "The ChaosIQ API service looks busy. Try again soon!")
    except requests.ReadTimeout as t:
        logger.warn(
            "The ChaosIQ API service took long to respond. Try again soon!")
    except requests.HTTPError as x:
        logger.debug(str(x))
        logger.error(
            "Something went wrong. Please contact ChaosIQ.")
    else:
        if r.status_code != expected_status:
            logger.debug(r.text)
            return

        return r



# -*- coding: utf-8 -*-
import json
import os.path
from typing import List

import click
from logzero import logger
import requests

from chaosiq import __version__
from chaosiq.config import initialize_config, load_config, \
    CHAOISQ_CONFIG_PATH, get_token, save_config, set_token, set_config_key
from chaosiq.exceptions import ChaosIQException, ConfigurationMissingError

__all__ = ["cli"]


###############################################################################
# chaosiq CLI
###############################################################################
@click.group()
@click.option('--config-file', default=CHAOISQ_CONFIG_PATH,
              show_default=True, help='chaosiq configuration file path.')
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx: click.Context, config_file: str=CHAOISQ_CONFIG_PATH):
    """
    The chaosiq command adds ChaosIQ features to the open-source chaostoolkit
    package.

    \b
    In order to benefit from thos features, you must have registered with
    ChaosIQ to get an access token. You should set that token in the
    configuration file with `chaosiq login`.
    \b
    Once you have configured chaosiq, simply use the chaostoolkit commands
    as usual, the chaosiq package will extend them by talking to the ChaosIQ
    services.
    """
    subcommand = ctx.invoked_subcommand
    ctx.obj = {}
    ctx.obj["config_file"] = config_file

    if subcommand == "config":
        return

    if not os.path.exists(config_file):
        click.echo(
            "You are missing a configuration file, creating '{c}'".format(
                c=config_file))
        initialize_config(config_file)

    try:
        ctx.obj["config"] = load_config(config_file)
        if subcommand != "login":
            ctx.obj["token"] = get_token(ctx.obj["config"])
    except ChaosIQException as err:
        raise click.ClickException(str(err))


@cli.command()
@click.pass_context
def login(ctx: click.Context):
    """
    Login with the chaosiq services.
    """
    email = click.prompt(click.style("Email", fg='green'), type=str)
    token = click.prompt(
        click.style("Token", fg='green'), type=str, hide_input=True)

    config = ctx.obj["config"]
    config_file = ctx.obj["config_file"]
    set_token(email, token, config)
    save_config(config, config_path=config_file)
    click.echo(
        "Credentials saved in configuration '{c}'".format(c=config_file))


@cli.command()
@click.option('--force', is_flag=True, help='Force the operation.')
@click.option('--key', help='Dotted key to set in the config.')
@click.option('--value', help='Value for a key.')
@click.argument("operation", type=click.Choice(["init", "set"]), nargs=1)
@click.pass_context
def config(ctx: click.Context, operation: str=None, key: str=None,
           value: str=None, force: bool=False):
    """
    Manage your chaosiq configuration.

    \b
    The following operations are supported:
    - init: to create a new default configuration (overriding the existing one)
    - set: to set a key in the configuration

    \b
    When setting a key, it can be dotted. In that case it is translated into
    nested keys inside the configuration. The value is always set to the last
    segment of the key.
    """
    config_file = ctx.obj["config_file"]
    if operation == "init":
        initialize_config(config_file, force)
        click.echo("The configuration was created at: {c}".format(
            c=config_file))
        return

    try:
        config = load_config(config_file)
    except ConfigurationMissingError as err:
        raise click.ClickException(str(err))

    if operation == "set":
        config = load_config(config_file)
        set_config_key(config, key, value)
        save_config(config, config_path=config_file)



# -*- coding: utf-8 -*-
import os
import os.path

import yaml

from chaosiq.exceptions import ConfigurationMissingError, \
    ConfigurationTokenMissing
from chaosiq.types import Configuration

__all__ = ["initialize_config", "load_config", "CHAOISQ_CONFIG_PATH",
           "DEFAULT_TOKEN", "get_token", "set_token", "set_config_key"]

CHAOISQ_CONFIG_PATH = os.path.abspath(os.path.expanduser("~/.chaosiq/config"))
DEFAULT_TOKEN = "<YOUR TOKEN>"
DEFAULT_CONFIG = {
    "auth": {
        "token": DEFAULT_TOKEN
    }
}


def load_config(config_path: str = CHAOISQ_CONFIG_PATH) -> Configuration:
    """
    Load the chaosiq configuration into memory. Raises
    `ConfigurationMissingError` when the config does not exist.
    """
    if not os.path.exists(config_path):
        raise ConfigurationMissingError(
            "Your chaosiq configuration is missing or is empty. "
            "Run `chaosiq config init` to initialize a default one.")

    with open(config_path) as f:
        return yaml.load(f.read())


def save_config(config: Configuration, config_path: str = CHAOISQ_CONFIG_PATH):
    """
    Save the config on to disk.
    """
    with open(config_path, 'w') as f:
        return f.write(yaml.dump(config, default_flow_style=False))


def initialize_config(config_path: str = CHAOISQ_CONFIG_PATH,
                      force: bool = False):
    """
    Initialize a default configuration on disk. If the cconfiguration already
    exists, it is left untouched, unless the `force` parameter is set to
    `True`.

    The configuration is set to be read and written only by the current user.
    """
    if os.path.exists(config_path) and not force:
        return

    try:
        os.mkdir(os.path.dirname(config_path))
    except FileExistsError:
        pass

    with open(config_path, "w") as f:
        os.chmod(config_path, 0o600)
        f.write(yaml.dump(DEFAULT_CONFIG, default_flow_style=False))


def set_token(email: str, token: str, config: Configuration,
              force: bool = False):
    """
    Save token to the configuration file.
    """
    config["auth"]["email"] = email
    config["auth"]["token"] = token


def get_token(config: Configuration) -> str:
    """
    Get the chaosiq access token from the config and raises errors when it
    is missing or unset.
    """
    token = config.get("auth", {}).get("token")
    if not token:
        raise ConfigurationTokenMissing(
            "Missing chaosiq token, please ensure you "
            "set it in your chaosiq configuration with `chaosiq login`.")

    if token == DEFAULT_TOKEN:
        raise ConfigurationTokenMissing(
            "It seems you have not set your chaosiq token "
            "in the configuration. Please, run `chaosiq login`.")

    return token


def set_config_key(config: Configuration, key: str, value: str):
    """
    Set a value to a key in the configuration. The key may be dotted to
    represent a nested key, the value is set to the last segment of the key.
    """
    tree = {}
    keys = key.split(".")

    c = config
    for k in keys[:-1]:
        if k not in c:
            c[k] = {}
        c = c[k]

    c[keys[-1]] = value



# -*- coding: utf-8 -*-
import io
import json
import os
import os.path
import re
import textwrap
from typing import Dict, List

from chaoslib.discovery import discover as disco
from chaoslib.exceptions import DiscoveryFailed
from chaoslib.types import Activity, Discovery, Experiment
from chaostoolkit.cli import discover as chtk_discover
from chaostoolkit.cli import init as chtk_init
from chaostoolkit.cli import run as chtk_run
import click
from logzero import logger
import requests

from chaosiq.api import call_chaosiq_api
from chaosiq.config import load_config, CHAOISQ_CONFIG_PATH
from chaosiq.exceptions import ConfigurationMissingError

__all__ = ["discover", "init", "run"]
CandidateHashRe = re.compile(r"\{\{([a-z0-9]{56,56})\}\}")


###############################################################################
# chaostoolkit CLI overloading
###############################################################################
@click.command(help=chtk_discover.__doc__)
@click.option('--no-system-info', is_flag=True,
              help='Do not discover system information.')
@click.option('--no-install', is_flag=True,
              help='Assume package already in PYTHONPATH.')
@click.option('--discovery-report-path', default="./discovery.json",
              help='Path where to save the report from the discovery.',
              show_default=True)
@click.argument('package')
@click.pass_context
def discover(ctx: click.Context, package: str,
             discovery_report_path: str = "./discovery.json",
             no_system_info: bool = False, no_install: bool = False):
    # invoking actual chaostoolkit parent discovery function
    try:
        discovery = disco(
            package_name=package, discover_system=not no_system_info,
            download_and_install=not no_install)
    except DiscoveryFailed as err:
        logger.debug("Failed to discover {}".format(package), exc_info=err)
        click.echo(str(err), err=True, nl=True)
        return

    try:
        config_file = os.environ.get(
            "CHAOISQ_CONFIG_PATH", CHAOISQ_CONFIG_PATH)
        config = load_config(click.format_filename(config_file))
    except ConfigurationMissingError as err:
        display_missing_config_message(config_file)
    else:
        logger.info("Calling chaosiq to fetch potential experiments")

        r = call_chaosiq_api(
            "/v1/flow/discovery/", payload=discovery, config=config,
            expected_status=201)

        if r:
            discovery = r.json()
            count = len(discovery.get("experiments", []))
            m = "{d} experiment suggestions were found"
            if count == 1:
                m = "{d} experiment suggestion was found"
            logger.info(m.format(d=count))

    with open(discovery_report_path, "w") as d:
        d.write(json.dumps(discovery, indent=2))

    logger.info("Discovery report saved in {p}".format(
        p=discovery_report_path))

    return discovery


@click.command(help=chtk_init.__doc__)
@click.option('--discovery-path', default="./discovery.json",
              help='Path to the discovery outcome.',
              show_default=True, type=click.Path(exists=False))
@click.option('--experiment-path', default="./experiment.json",
              help='Path where to save the experiment.',
              show_default=True)
@click.pass_context
def init(ctx: click.Context, discovery_path: str = "./discovery.json",
         experiment_path: str = "./experiment.json"):
    config = None
    choice = "default"
    try:
        config_file = os.environ.get(
            "CHAOISQ_CONFIG_PATH", CHAOISQ_CONFIG_PATH)
        config = load_config(click.format_filename(config_file))
    except ConfigurationMissingError as err:
        display_missing_config_message(config_file)
    else:
        click.secho(
            "\nYou have installed `chaosiq` and can take the opportunity to\n"
            "let ChaosIQ guide you through creating rich chaos experiments\n"
            "tailored to your system.", fg="blue")

        choice = click.prompt(
            click.style(
                "Use 'chaosiq' init feature or the chaostoolkit 'default'",
                dim=True),
            default="default", show_default=True,
            type=click.Choice(['default', 'chaosiq']))

    if choice == "default":
        experiment = ctx.forward(chtk_init)

        if config:
            call_chaosiq_api(
                "/v1/flow/init/", payload=experiment, config=config,
                expected_status=201)
        return

    if not os.path.exists(discovery_path):
        click.secho(
            "No discovery output found.\n\n"
            "Are you running from a directory where the discovery\n"
            "output, usually `discovery.json`, is located? You can\n"
            "specify the path to that file with `--discovery-path`.\n"
            "Alternatively, have you run `chaos discover` already?\n",
            err=True)
        return

    discovery = None
    with open(discovery_path) as d:
        discovery = json.loads(d.read())

    suggested_experiments = discovery.get("experiments", [])
    if not suggested_experiments:
        click.secho(
            "No experiments found in the discovery output.\n\n"
            "Have you run the `chaos discover command yet?\n", err=True)
        return

    experiments = []
    for experiment in suggested_experiments:
        experiments.append((
            experiment["title"],
            experiment
        ))

    experiment = select_experiment(experiments)
    m = "\nYou have selected the experiment: '{t}'".format(
        t=experiment["title"])
    logger.debug(m)
    click.secho(m, fg="magenta")

    desc = experiment.get("description")
    if desc:
        click.secho("\n".join(textwrap.wrap(desc, width=70)), dim=True)

    set_experiment_arguments(discovery, experiment)

    r = call_chaosiq_api(
        "/v1/flow/init/", payload=experiment, config=config,
        expected_status=201)
    if r:
        experiment = r.json()

    m = "\nSaving your experiment in '{e}'".format(e=experiment_path)
    logger.debug(m)
    click.secho(m)
    with open(experiment_path, "w") as e:
        experiment = sort_experiment_keys(experiment)
        e.write(json.dumps(experiment, indent=4, sort_keys=False))

    m = "\nYour experiment is now completed and saved. In some cases, this\n"\
        "is enough to go and run via `chaos run {}`. However,\n"\
        "we suggest you open the experiment and review it before executing\n"\
        "it.".format(experiment_path)
    logger.debug(m)
    click.secho(m, fg="blue")

    return experiment


@click.command(help=chtk_run.__doc__)
@click.option('--journal-path', default="./journal.json",
              help='Path where to save the journal from the execution.')
@click.option('--dry', is_flag=True,
              help='Run the experiment without executing activities.')
@click.option('--no-validation', is_flag=True,
              help='Do not validate the experiment before running.')
@click.argument('path', type=click.Path(exists=True))
@click.pass_context
def run(ctx: click.Context, path: str, journal_path: str="./journal.json",
        dry: bool=False, no_validation: bool=False):
    journal = ctx.forward(chtk_run)

    try:
        config_file = os.environ.get(
            "CHAOISQ_CONFIG_PATH", CHAOISQ_CONFIG_PATH)
        config = load_config(click.format_filename(config_file))
    except ConfigurationMissingError as err:
        display_missing_config_message(config_file)
    else:
        r = call_chaosiq_api(
            "/v1/flow/run/", payload=journal, config=config,
            expected_status=201)
        if r:
            journal = r.json()
            with io.open(journal_path, "w") as r:
                json.dump(journal, r, indent=2, ensure_ascii=False)

    return journal


###############################################################################
# Private functions
###############################################################################
def select_experiment(experiments: List[Experiment]) -> Experiment:
    """
    Prompt the user to select an experiment from the list of experiments
    found in the discovery.
    """
    click.secho(
        "\nWith the discovery of your system capabilities, ChaosIQ\n"
        "correlated relevant experiments with your system. You may now\n"
        "select one of those returned experiments to learn from your system\n"
        "under stress.\n"
        "Select now one of the experiments suggested by ChaosIQ:\n",
        fg="blue")

    echo = click.echo
    if len(experiments) > 10:
        echo = click.echo_via_pager

    echo("\n".join([
        "{i}) {t}".format(
            i=idx+1, t=title) for (idx, (title, exp)) in enumerate(
                experiments)]))
    index = click.prompt(
        click.style('\nPlease select an experiment', dim=True), type=int)

    index = index - 1
    if index > len(experiments):
        click.secho("This is not a valid experiment", fg="yellow")
        return select_experiment(experiments)

    experiment = experiments[index]
    return experiment[1]


def display_missing_config_message(config_file: str):
    """
    Log a user-friendly and actionable message when chaosiq was not configured.
    """
    logger.debug("chaosiq not configured")
    click.secho("\nYou have installed the `chaosiq` package but you\n"
                "don't seem to have configured it.\n\n"
                "To benefit from ChaosIQ, you must have registered\n"
                "with http://chaosiq.io and used the `chaosiq login`\n"
                "command to store your token in '{}'.\n\n"
                "If you do not intend on using the ChaosIQ services,\n"
                "you should uninstall the package with\n"
                "`pip uninstall chaosiq` and run this command again.\n\n"
                "We will continue with the default chaostoolkit behavior.\n"
                .format(config_file), dim=True, err=True)


def set_experiment_arguments(discovery: Discovery, experiment: Experiment):
    """
    Iterate through all activities arguments and prompt the user for a choice
    whenever we gathered a candidate during discovery.
    """
    candidates = discovery.get("candidates")
    if not candidates:
        click.secho(
            "This experiment could not be matched against your system's\n"
            "state. This either means that none of the activities of this\n"
            "experiment expect your input. Or none could be set against any\n"
            "the discovered aspects of your system. Please, open the\n"
            "experiment file once this command finishes.\n", fg="blue")
        return

    exp_ref = experiment["chaosiq"]["suggested_experiment_ref"]
    if exp_ref in candidates:
        candidates = candidates[exp_ref]
        click.secho(
            "\nThis experiment contains activities which arguments were\n"
            "matched with candidates from your system by ChaosIQ. You can\n"
            "now pick the appropriate value for each argument now.",
            fg="blue")
        q = "Do you wish to set arguments from your discovered system?"

        if click.confirm(click.style(q, dim=True)):
            click.echo("")
            hypo = experiment.get("steady-state-hypothesis", {})
            probes = hypo.get("probes", [])
            if requires_filling(probes):
                click.secho(
                    "Fill the steady-state hypothesis probe arguments:")
                set_activity_arguments(probes, candidates)

            method = experiment.get("method", [])
            if requires_filling(method):
                click.secho("Fill the method activities arguments:")
                set_activity_arguments(method, candidates)

            rollbacks = experiment.get("rollbacks", [])
            if requires_filling(rollbacks):
                click.secho("Fill the rollbacks arguments:")
                set_activity_arguments(rollbacks, candidates)

            click.secho(
                "\nYou have now set all activities arguments!", fg="blue")

        replace_not_matched_arguments(experiment, candidates)


def replace_not_matched_arguments(experiment: Experiment,
                                  candidates: Dict[str, str]):
    """
    In case the experiment has remaining hashes that were not matched with a
    candidate, we replace those tokens with empty striings as friendlier to
    the user.
    """
    hypo = experiment.get("steady-state-hypothesis", {})
    activities = []
    activities.extend(hypo.get("probes", []))
    activities.extend(experiment.get("method", []))
    activities.extend(experiment.get("rollbacks", []))

    for activity in activities:
        arguments = activity.get("provider", {}).get("arguments", {})
        if arguments:
            for k, v in arguments.items():
                for m in CandidateHashRe.finditer(v):
                    hsh = m.group(1)
                    if hsh not in candidates:
                        v = v.replace(m.group(0), "")
                arguments[k] = v


def requires_filling(activities: List[Activity]):
    """
    Go through all activities and return `True` as soon as one of them
    requires the user to select a value from the candidates we discovered.
    """
    for activity in activities:
        arguments = activity.get("provider", {}).get("arguments", {})
        if arguments:
            for k, v in arguments.items():
                if CandidateHashRe.match(v):
                    return True
    return False


def set_activity_arguments(activities: List[Activity],
                           candidates: Dict[str, str]):
    """
    Prompt the user for argument values from the discovered candidates.
    """
    for activity in activities:
        arguments = activity.get("provider", {}).get("arguments", {})
        if not arguments:
            continue
        click.secho("\nActivity: {}".format(activity["name"]), fg="green")
        for (arg_key, arg_value) in arguments.items():
            click.secho(" Argument: {}".format(arg_key), fg="yellow")
            for m in CandidateHashRe.finditer(arg_value):
                hsh = m.group(1)
                if hsh in candidates:
                    values = candidates[hsh]
                    q = "Please, pick one of the following value "\
                        "for this argument"

                    click.secho("\n".join([
                        " {i}) {t}".format(
                            i=i+1,
                            t=v) for (i, v) in enumerate(values)]))
                    index = click.prompt(click.style(
                            'Select a value (0 to leave empty)',
                            dim=True), type=int)

                    index = index - 1
                    if index > len(values):
                        click.secho("Invalid choice", fg="red")
                        continue
                    elif index == -1:
                        arg_value = arg_value.replace(m.group(0), "")
                    else:
                        arg_value = arg_value.replace(
                            m.group(0), values[index])

            arguments[arg_key] = arg_value
            click.echo("")


def sort_experiment_keys(experiment: Experiment) -> Experiment:
    """
    Ensure the experiment keys are in the right order for use-friendlyness.
    The toolkit doesn't care either way but it makes the experiment much more
    readable and comprehensible.

    This works because in recent Python versions, insertion ordering is kept in
    a dictionary.
    """
    exp = {}
    if "version" in experiment:
        exp["version"] = experiment["version"]

    if "title" in experiment:
        exp["title"] = experiment["title"]

    if "description" in experiment:
        exp["description"] = experiment["description"]

    if "tags" in experiment:
        exp["tags"] = experiment["tags"]

    if "chaosiq" in experiment:
        exp["chaosiq"] = experiment["chaosiq"]

    if "configuration" in experiment:
        exp["chaosiq"] = experiment["configuration"]

    if "secrets" in experiment:
        exp["secrets"] = experiment["configuration"]

    if "steady-state-hypothesis" in experiment:
        exp["steady-state-hypothesis"] = experiment["steady-state-hypothesis"]

    if "method" in experiment:
        exp["method"] = experiment["method"]

    if "rollbacks" in experiment:
        exp["rollbacks"] = experiment["rollbacks"]

    return exp



# -*- coding: utf-8 -*-

__all__ = ["ChaosIQException", "ConfigurationMissingError",
           "ConfigurationTokenMissing"]


class ChaosIQException(Exception):
    pass


class ConfigurationMissingError(ChaosIQException):
    pass


class ConfigurationTokenMissing(ChaosIQException):
    pass
